use crate::vm::DmValue::*;
use crate::vm::{mem, DmValue, DmVirtualMachine};

pub fn dm_print(args: Vec<DmValue>, vm: &mut DmVirtualMachine) -> DmValue {
    if args.len() != 1 {
        return DmUnit // TODO: make exception
    }
    match &args[0] {
        DmByte(b) => {
            print!("{}", *b);
        }
        DmInt(i) => {
            print!("{}", *i);
        }
        DmLong(l) => {
            print!("{}", *l);
        }
        DmDouble(d) => {
            print!("{}", *d);
        }
        DmBoolean(b) => {
            print!("{}", *b);
        }
        DmPointer(ptr) => unsafe {
            let dm_alloc_object = Box::from_raw(*ptr);
            let dm_object_type = &dm_alloc_object.object_type;
            if let Some(to_string_method) = dm_object_type.get_method(String::from("std::core::Object::to_string"), &dm_alloc_object) {
                let call_result = vm.call(to_string_method, vec![args[0].clone()]);
                if let DmPointer(dm_string_pointer) = call_result {
                    let dm_string_object = Box::from_raw(dm_string_pointer);
                    let bytes_property = dm_string_object.object_type.get_property(String::from("bytes"), &dm_string_object).unwrap();
                    let string_bytes_value = mem::get_property_value(bytes_property, &dm_string_object);
                    if let DmByteArray(raw_string_bytes) = string_bytes_value {
                        // TODO: other encodings
                        print!("{}", String::from_utf8(raw_string_bytes).unwrap());
                    } else {
                        panic!("Expected std::core::String.bytes to be a ByteArray.")
                    }
                } else {
                    panic!("Expected {} to return a string.", to_string_method.fqn)
                }
            } else {
                println!("{}@{:?}", dm_object_type.get_fqn(), dm_alloc_object.data)
            }
        },
        DmByteArray(bs) => {

        }
        DmIntArray(is) => {

        }
        DmLongArray(ls) => {}
        DmDoubleArray(ds) => {}
        DmBooleanArray(bs) => {}
        DmPointerArray(ptrs) => {}
        DmUnit => {
            print!("Unit")
        },
    }
    DmUnit
}
